#!/usr/bin/env python
# -*- coding: utf-8 -*-

#    Magic Desk Compatible Cartridge Generator (c) 2013  Žarko Živanov
#    Cartridge schematics and PCB design (c) 2013 Marko Šolajić
#    E-mails: zzarko and msolajic at gmail

#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.

#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
from __future__ import division

VERSION="1.0b"

import copy
import array
import os
import glob
import sys
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
import re

#help function for debug
def print_matrix(matrix,name=""):
    height = len(matrix)
    width = len(matrix[0])
    print(name)
    for y in range(height):
        for x in range(width):
            print("%02x " % matrix[y][x], end='')
        print()

#commodore 64 screen codes
SCR_CODES="""@abcdefghijklmnopqrstuvwxyz[|]|| !"#$%&'()*+,-./0123456789:;<=>?|ABCDEFGHIJKLMNOPQRSTUVWXYZ"""
SCR_DICT={SCR_CODES[pos]:pos for pos in range(len(SCR_CODES))}
#commodore 64 petascii codes
PETASCII_CODES="""0123456789|||||||abcdefghijklmnopqrstuvwxyz||||<|ABCDEFGHIJKLMNOPQRSTUVWXYZ"""
PETASCII_DICT={PETASCII_CODES[pos]:pos+48 for pos in range(len(PETASCII_CODES))}
#keys used to select menu items (max 62 programs plus Basic)
KEYS="1234567890abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

#cartridge size (64/128/256k)
cart_size = 128*1024
#directory for prg files
prg_directory  = "prg"
#cartridge menu program
cart_menu_file = "menu"
#finished cartridge bin file
cart_file_name = "compilation.bin"
#order of programs in CFG file
cart_order = []
#border color after program select
color_border = 14
#background color after program select
color_background = 6
#character color after program select
color_character = 14
#default menu title
menu_title = ""

#returns array of screen codes
def to_screen(s):
    return [SCR_DICT[x] for x in s]

#returns the size of an open file
def flen(f):
    return os.fstat(f.fileno()).st_size

#adds one byte to array
def append_byte(arr,byte):
    arr.append(byte & 0xff)

#adds two bytes to array (lo/hi)
def append_word(arr,word):
    arr.append(word & 0xff)
    arr.append(word >> 0x08)

def usage(code=0):
    print("""Usage:
python {0}
 - scans the prg directory and places all prg files on cartridge
python {0} <cfg_file> (with or without extension)
 - reads cartridge configuration from CFG file

All programs for cartridge must be in prg format!

Create prg directory if it doesn't exist and place all prg files
for cartridge inside. To make specific program order, you can add
NN_ prefix to prg files (N - 0-9). This prefix will be excluded
from name in menu. Or, you can use CFG file to define cartridge.
Sample CFG file is provided with installation.

Generated bin file can be burned to Magic Desk compatibile
eprom/cartridge, or converted to crt with VICE's cartconv:
cartconv -t md -i <bin_file> -o <crt_file>
""".format(sys.argv[0]))
    exit(code)

def error(s):
    print("\nERROR!! {0}\n".format(s))
    usage(1)

print("Magic Desk Cartridge Generator v%s" % VERSION)

#matrices for character and color screen
screen = [[0xa0 for col in range(40)] for row in range(25)]
colors = [[0x06 for col in range(40)] for row in range(25)]

items = []

if len(sys.argv) > 2: usage()
if len(sys.argv) > 1:
    #read configuration from cfg file
    cfgfile = sys.argv[1]
    if cfgfile[-4:] != ".cfg" and cfgfile[-4:] != ".CFG":
        cfgfile += ".cfg"
    if os.access(cfgfile, os.F_OK):
        cfg = configparser.SafeConfigParser()
        cfg.read(cfgfile)
        programs = cfg.sections()
        if cfg.has_section("cartridge"):
            if cfg.has_option("cartridge","size"):
                cart_size = cfg.getint("cartridge","size")
                if cart_size not in [64,128,256,512]:
                    error("Wrong cartridge size, must be 64,128,256 or 512")
                cart_size *= 1024
            if cfg.has_option("cartridge","order"):
                cart_order = cfg.get("cartridge","order")
                cart_order = cart_order.split(",")
            if cfg.has_option("cartridge","name"):
                cart_file_name = cfg.get("cartridge","name")
                if cart_file_name[-4:] != ".bin" and cart_file_name[-4:] != ".BIN":
                    cart_file_name += ".bin"
            if cfg.has_option("cartridge","menu"):
                cart_menu_file = cfg.get("cartridge","menu")
                if cart_menu_file[-4:] == ".prg" or cart_menu_file[-4:] == ".PRG":
                    cart_menu_file = cart_menu_file[:-4]
            if cfg.has_option("cartridge","border"):
                color_border = cfg.getint("cartridge","border")
                if color_border not in range(16):
                    error("Border color must be 0-15")
            if cfg.has_option("cartridge","background"):
                color_background = cfg.getint("cartridge","background")
                if color_background not in range(16):
                    error("Background color must be 0-15")
            if cfg.has_option("cartridge","character"):
                color_character = cfg.getint("cartridge","character")
                if color_character not in range(16):
                    error("Character color must be 0-15")
            if cfg.has_option("cartridge","title"):
                menu_title = cfg.get("cartridge","title")
                if len(menu_title) > 40:
                    error("Menu title must be 40 characters or less")
            programs.remove("cartridge")
        if len(cart_order) == 0:
            cart_order = ["{:02d}".format(i) for i in range(64)]
        for prgno in cart_order:
            prgsec = "prg"+prgno
            if not cfg.has_section(prgsec): continue
            prgname = cfg.get(prgsec,"file")
            if prgname[-4:] == ".prg" or prgname[-4:] == ".PRG":
                prgname = prgname[:-4]
            if cfg.has_option(prgsec,"name"):
                name = cfg.get(prgsec,"name")
            else:
                name = prgname[:-4].replace("_"," ")
            if cfg.has_option(prgsec,"run"):
                runaddr = cfg.get(prgsec,"run")
                if runaddr[0] == '$': runaddr = runaddr.replace("$","0x")
                runaddr = int(runaddr,0)
            else:
                runaddr = 0
            items.append([name,prgname,runaddr,0,0,0])
    else:
        error("CFG file '{0}' not found.".format(cfgfile))
else:
    #collect all prg files in prg directory
    prgList = glob.glob( os.path.join(prg_directory, '*.[pP][rR][gG]') )
    if len(prgList) == 0:
        error("No prg files found. Make sure to place them in prg directory.\n")
    prefix = re.compile("\A[0-9][0-9]_")
    prgList.sort()
    for prg in prgList:
        prgname = os.path.basename(prg)[:-4]
        print(prgname)
        nstart = 3 if prefix.match(prgname) else 0
        name = prgname[nstart:].replace("_"," ")
        items.append([name,prgname,0,0,0,0])

item_basic = ["Basic","",0xfce2,0,0,0]  #always added at the end
items.append(item_basic)
item_len = min(max(len(x[0]) for x in items),30)    #menu name is max 30 characters

#center one line of text on specified row
def screen_center(row,text):
    l = len(text)
    pos = (40 - l) // 2
    text = to_screen(text)
    text = [i+128 for i in text]
    screen[row][pos:pos+len(text)] = text

#menu characters and colors
menu_width = item_len+7             #longest name plus borders and key representation
max_screen_items = 10               #max number of items displayed in menu
if len(items) <= max_screen_items:
    menu_height = len(items)*2+2
    max_screen_items = len(items)
else:
    menu_height = max_screen_items*2+2
    #crsr usage message
#    item_text = to_screen("Use CRSR to scroll programs.")
#    item_text = [i+128 for i in item_text]
#    screen[24][6:6+len(item_text)] = item_text
    screen_center(24,"Use CRSR to scroll programs.")
menu_scr = [[0x20 for col in range(menu_width)] for row in range(menu_height)]
menu_col = [[0x05 for col in range(menu_width)] for row in range(menu_height)]

#menu item definitions
menu_item_scr = [ [0xa0,0xb1,0xa0,0x20],[0x20]*item_len ]
menu_item_col = [ [0x0f,0x0f,0x0f,0x05],[0x05]*item_len ]

#create menu border
for x in range(1,menu_width-1):
    menu_scr[0][x] = 0xe2
    menu_scr[-2][x] = 0x62
    menu_scr[-1][x] = 0xe6
    menu_col[0][x] = 0x0a
    menu_col[-2][x] = 0x0a
    menu_col[-1][x] = 0x02
for y in range(1,menu_height-1):
    menu_scr[y][0] = 0x61
    menu_scr[y][-2] = 0xe1
    menu_scr[y][-1] = 0xe6
    menu_col[y][0] = 0x0a
    menu_col[y][-2] = 0x0a
    menu_col[y][-1] = 0x02
menu_scr[0][0] = 0xec
menu_scr[0][menu_width-2] = 0xfb
menu_scr[menu_height-2][0] = 0xfc
menu_scr[menu_height-2][menu_width-2] = 0xfe
menu_scr[-1][-1] = 0xe6
menu_scr[-1][0] = 0xa0
menu_scr[0][-1] = 0xa0
menu_col[0][0] = 0x0a
menu_col[0][menu_width-2] = 0x0a
menu_col[menu_height-2][0] = 0x0a
menu_col[menu_height-2][menu_width-2] = 0x0a
menu_col[-1][-1] = 0x02
menu_col[-1][0] = 0x06
menu_col[0][-1] = 0x06
#print_matrix(menu_scr,"menu_scr")
#print_matrix(menu_col,"menu_col")

#fill menu item colors
for i in range(max_screen_items):
    if i == 0:
        itemy = 1
    else:
        itemy += 2
    item_col = copy.deepcopy(menu_item_col)
    menu_col[itemy][1:5] = item_col[0]
    menu_col[itemy][5:5+item_len] = item_col[1]
#print_matrix(menu_col,"menu_col")

#place menu title on screen, if defined
if menu_title != "":
    screen_center(0,menu_title)
    move_menuy = 1
else:
    move_menuy = 0

#insert menu in screen
menux = (40-menu_width) // 2
menuy = (25-menu_height) // 2 + move_menuy
for i,(chars,cols) in enumerate(zip(menu_scr,menu_col)):
    screen[menuy+i][menux:menux+menu_width] = chars
    colors[menuy+i][menux:menux+menu_width] = cols
menu_line = copy.deepcopy(screen[menuy+2])          #40 char line for menu entries
#print_matrix(menu_scr,"menu_scr")
#print_matrix(menu_col,"menu_col")

#generate menu items
itemx = menux + 1
for i,item in enumerate(items):
    item_scr = copy.deepcopy(menu_item_scr)
    #fill menu entry key
    if item[1] != "":
        item_scr[0][1] = SCR_DICT[KEYS[i]]+128      #inverse char
    else:
        item_scr[0][1] = 0x9f                       # <- (Basic)
    #fill menu entry text
    item_text = to_screen(item[0][:30])
    item_scr[1][:len(item_text)] = item_text
    #create 40 char menu line
    new_menu_line = copy.deepcopy(menu_line)
    new_menu_line[itemx:itemx+4] = item_scr[0]
    new_menu_line[itemx+4:itemx+4+item_len] = item_scr[1]
    item.append(new_menu_line)
    #print(new_menu_line)

#load all programs
item_no = 0
for i,item in enumerate(items):
    if item[1] == "": continue
    prgfile=os.path.join(prg_directory,item[1])
    if os.access(prgfile+".PRG", os.F_OK):
        prgfile += ".PRG"
    else:
        prgfile += ".prg"
    prg = open(prgfile,"rb")
    temp = array.array('B')
    temp.fromfile(prg,flen(prg))
    addr = temp.pop(0) + 256*temp.pop(0)
    if item[2] == 0:        #if start address
        item[2] = addr
    item[5] = temp.buffer_info()[1]
    item.append(temp)
    item_no += 1
if item_no > len(KEYS):
    error("Cartridge can have max %d programs, %d too many supplied."
          % (len(KEYS), item_no-len(KEYS)))

#open cart menu file
if os.access(cart_menu_file+".PRG", os.F_OK):
    cart_menu_file += ".PRG"
else:
    if os.access(cart_menu_file+".prg", os.F_OK):
        cart_menu_file += ".prg"
    else:
        error("cartridge menu program '%s' not found." % cart_menu_file)
cart_prg = open(cart_menu_file,"rb")
#skip start address
cart_prg.seek(2)
cart_file = array.array('B')
cart_file.fromfile(cart_prg,flen(cart_prg)-2)
cart_prg.close()

#calculate table addresses
table_data = array.array('B')
bank = 0
#49 = 9 bytes for prgtable, 40 for item screen characters
address = cart_file.buffer_info()[1] + len(items)*49 + 0x8000
crtaddress = address - 0x8000
tbladdress =  cart_file.buffer_info()[1] + len(items)*40 + 0x8000
print("Cartridge %dk memory map:\n%20s located at $%06x" % (cart_size//1024, "menu code", 0))
for i,item in enumerate(items):
    append_byte(table_data,bank)        #bank, 1 byte
    append_word(table_data,address)     #address in bank, 2 bytes
    if item[1] != "":
        print("%20s located at $%06x, run address:" % (item[1],crtaddress), end=" ")
        length = item[7].buffer_info()[1]
        append_word(table_data,length)  #length, 2 bytes
    else:
        append_word(table_data,0)
    append_word(table_data,item[2])     #start address, 2 bytes
    if item[2] == 2049:
        append_word(table_data,0)       #run address, 2 bytes (0 - BASIC RUN)
        print("BASIC RUN")
    else:
        append_word(table_data,item[2])
        if item[1] != "":
            print("$%04x" % item[2])
    address += length
    crtaddress += length
    while address > 0x9fff:
        address -= 0x2000
        bank += 1

#remove screen, colors, keys and variables at the end
l = 0x400+0x400+64+5+2+3
del cart_file[-l:]

#assemble cartridge
#for details about various fields in here, check C64 assembler source
append_byte(cart_file,color_border)         #border color
append_byte(cart_file,color_background)     #background color
append_byte(cart_file,color_character)      #character color
append_word(cart_file,(menuy+1)*40+0x400)   #screen address of first menu item
append_byte(cart_file,max_screen_items)     #max number of options in menu
append_byte(cart_file,len(items)-max_screen_items)         #max index of first menu item
append_byte(cart_file,len(items))           #total number of options/programs
keys = array.array('B',[0]*64)              #valid keys for menu
for i,item in enumerate(items):
    if item[5] > 0:
        keys[i] = PETASCII_DICT[KEYS[i]]
    else:
        keys[i] = 95                        # <- (Basic)
cart_file.extend(keys)
for y in range(25):                         #screen 40x25 + 24 bytes padding
    for x in range(40):
        append_byte(cart_file,screen[y][x])
for i in range(24):
    append_byte(cart_file,0)
for y in range(25):                         #colors 40x25 + 24 bytes padding
    for x in range(40):
        append_byte(cart_file,colors[y][x])
for i in range(24):
    append_byte(cart_file,0)
append_word(cart_file,tbladdress)           #pointer to prgtable
for item in items:                          #menuitems
    for x in range(40):
        append_byte(cart_file,item[6][x])
cart_file.extend(table_data)                #prgtable
for i,item in enumerate(items):             #programs
    if item[1] == "": continue
    cart_file.extend(item[7])

#pad to cartridge size
length = cart_file.buffer_info()[1]
if length > cart_size:
    larger = length - cart_size
    error("Cartridge is %d bytes (%d blocks) larger than it should be."
          % (larger, larger // 254))
if length < cart_size:
    unused = cart_size - length
    temp = array.array('B',[0xff]*unused)
    cart_file.extend(temp)
    print("Unused %d bytes / %d block(s)" % (unused , unused // 254))

#write cartridge to file
cartridge = open(cart_file_name,"wb")
cart_file.tofile(cartridge)
cartridge.close()
print("Cartridge saved as '{0}'".format(cart_file_name))

